#!/bin/bash
################################################################################
# redirectModbus.sh
#
# Author: Piotr Golonka, CERN EN/ICE-SCD
# Date: 24 June 2014
#
# Usage:
#  1) [sudo] redirectModbus.sh -configure
#    -> reads the configuration from the default file /opt/unicryo/PVSS_projects/redirectModbus.cfg
#       keeping/modifying the current redirects based on the parameters from config file
#  2) [sudo] redirectModbus.sh -configure -cfg /opt/someuser/redirectModbus.cfg
#    -> reads the configuration from the specific config file /opt/someuser/redirectModbus.cfg
#       keeping/modifying the current redirects based on the parameters from config file
#  3) [sudo] redirectModbus.sh -configureClear
#    -> reads the configuration from /opt/unicryo/PVSS_projects/redirectModbus.cfg
#       clears all the current redirects and replaces with the ones from config file
#  4) [sudo] redirectModbus.sh -configureClear -cfg /opt/someuser/redirectModbus.cfg
#    -> reads the configuration from specific config file /opt/someuser/redirectModbus.cfg
#       clears all the current redirects and replaces with the ones from config file
#  5) [sudo] redirectModbus.sh -clear
#    -> clears all the redirects
#  6) [sudo] redirectModubus.sh -status
#     -> display the currently active redirection map
#  7) [sudo] redirectModubus.sh <PLC_IP> <WCCOAmod_port>
#     -> applies a single redirection (see below)
#
# Description:
# ------------
# This script reconfigures the Linux IP stack (iptables) to reroute the 
# packets for the PVSS MODBUS driver (port 502), so that many projects could run
# on the same machine.
#
# The scripts allow to configure the redirection of packets depending on the IP 
# address they are coming from. The packets are redirected to the same machine, 
# yet to a specified port.
#
# Such solution solves a problem of MODBUS in "push" mode, where the destination 
# TCP/IP port needs to be equal to 502. If we want to run multiple PVSS projects
# in the same machine, we need more instances of the MODBUS driver running,
# and each of them needs to listen on a different TCP/IP port. 
#
# The script needs to be executed by root or by a user with sudo rights
# (in the latter case, sudo may prompt for user's password).
#
# Note that the script is intelligent enough to see whether the redirection
# rule already exists for a specified PLC and
# adapts the settings appropriately
#
# Example: 
# --------
#  PLC1 (172.18.25.136) should be connected to WCCOAmod of Project1
#  PLC2 (172.18.37.9) should be connected to WCCOAmod of Project2
#
#  - configure WCCOAmod on Project1 to start with port number 8502
#   and on Project2 with port number 9502. You do it in the config file, eg
#   for Project1:
#	[mod_2]
#	tcpServerPort=8502
#
#  - to activate the redirection, execute
#	redirectModbus.sh 172.18.25.136 8502
#	redirectModbus.sh 172.18.37.9   9502
#
# For more details about the origins, look, for instance, at:
# https://icecontrols.cern.ch/jira/browse/ENS-1922
# https://icecontrols.cern.ch/jira/browse/ENS-3875
# look also at https://icecontrols.cern.ch/jira/browse/SCD-37
################################################################################

CFG_FILE='/opt/unicryo/PVSS_projects/redirectModbus.cfg'


function usage()
{
        echo "Usage (one of options):"
        echo " [sudo] $0 -configure "
        echo " [sudo] $0 -clear "
        echo " [sudo] $0 -status "
        echo " [sudo] $0 -configureClear "
        echo " [sudo] $0 <PLC_IP> <WCCOAmod_port>"
}


# Test an IP address for validity:
# http://www.linuxjournal.com/content/validating-ip-address-bash-script
# Usage:
#      valid_ip IP_ADDRESS
#      if [[ $? -eq 0 ]]; then echo good; else echo bad; fi
#   OR
#      if valid_ip IP_ADDRESS; then echo good; else echo bad; fi
#
function is_valid_ip()
{
    local  ip=$1
    local  stat=1

    if [[ $ip =~ ^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}$ ]]; then
        OIFS=$IFS
        IFS='.'
        ip=($ip)
        IFS=$OIFS
        [[ ${ip[0]} -le 255 && ${ip[1]} -le 255 \
            && ${ip[2]} -le 255 && ${ip[3]} -le 255 ]]
        stat=$?
    fi
    return $stat
}

# Do the redirection of a single port:
function redirectPort()
{
PLC_IP=$1
DEST_PORT=$2

# guess what is the IP address on which the packet comes by looking up the route to this IP
MY_IP=`/sbin/ip route get $PLC_IP |awk '/src/{print $NF;}'`

if [ -z "${MY_IP}" ]; then
    echo "Could not get the IP number of the port used to communicate with $PLC_IP"
    exit -1;
fi

## NOTE! We need to detect when we already run through sudo, or not...

IPTABLES_CMD='/sbin/iptables'
if [ $UID != 0 ]; then
    IPTABLES_CMD='sudo /sbin/iptables'
fi;

## Check if an entry already exist in the iptables
OUT=""
OUT=`$IPTABLES_CMD -t nat -n --line-numbers -L PREROUTING |awk -v plc_ip=${PLC_IP} -v dport=${DEST_PORT} '/DNAT/{split($9,a,"(.)*:"); if ($5==plc_ip) print $1,$5,a[2]};'`

if [ -z "${OUT}" ]; then
    # it does not exist in the table... add it
	$IPTABLES_CMD -t nat -A PREROUTING -p tcp -s ${PLC_IP} --dport 502 -d ${MY_IP} -j DNAT --to-destination $MY_IP:${DEST_PORT}
else
    # otherwise, replace the existing rule
    RULE_TO_REPLACE=`echo $OUT|awk '{print $1}'`;
    $IPTABLES_CMD -t nat -R PREROUTING ${RULE_TO_REPLACE} -p tcp -s ${PLC_IP} --dport 502 -d ${MY_IP} -j DNAT --to-destination $MY_IP:${DEST_PORT} 
fi;

}

function clearRedirects()
{

IPTABLES_CMD='/sbin/iptables'
if [ $UID != 0 ]; then
    IPTABLES_CMD='sudo /sbin/iptables'
fi;

# get the list of all redirects of port 502, get their positions and sort descending
OUT=`$IPTABLES_CMD -t nat -n --line-numbers -L PREROUTING |awk '/dpt:502/{print $1}'|sort -nr`

for rulenum in $OUT; do
    $IPTABLES_CMD -t nat -D PREROUTING $rulenum
done;
### Then clean them one by one

}


function showStatus()
{

IPTABLES_CMD='/sbin/iptables'
if [ $UID != 0 ]; then
    IPTABLES_CMD='sudo /sbin/iptables'
fi;

# get the list of all redirects of port 502, get their positions and sort descending
$IPTABLES_CMD -t nat -n --line-numbers -L PREROUTING |awk  '/DNAT/{split($9,a,"(.)*:"); print $5," => "a[2]};'
}

function redirectFromCfgFile()
{
    DO_CLEAR=""
    if [ $# = 1 ]; then DO_CLEAR=$1; fi;

    if [ ! -r $CFG_FILE ]; then
	echo "Error: cannot read config file $CFG_FILE"
	exit 1;
    fi
    
    line=0

    # Firstly dry-run the parsing to make sure we have correct parameters
    cat $CFG_FILE|while read PLC_IP PORT_NUMBER PROJECT_NAME DRIVER_NUMBER;
    do {
	line=$(($line + 1 ))

	if [ -z $PLC_IP ]; then continue; fi; #skip empty lines
	if [ x${PLC_IP:0:1} == "x#" ]; then continue; fi; #skip comment lines

	if ! is_valid_ip $PLC_IP; then 
	    echo "Error in line $line of config file: invalid PLC_IP string (got:$PLC_IP)"
	    exit 2;
	fi

	if [ -z $PORT_NUMBER ]; then 
	    echo "Error in line $line of config file: cannot parse PORT_NUMBER"
	    exit 3;
	fi;
	
	if ! [[ $PORT_NUMBER =~ ^[0-9]+$ ]]; then 
	    echo "Error in line $line of config file: PORT_NUMBER needs to be numeric (got:$PORT_NUMBER)"
	    exit 4;
	fi
	if [[ $PORT_NUMBER -gt 65535 || $PORT_NUMBER  -le 1024  ]]; then 
	    echo "Error in line $line of config file: PORT_NUMBER out or range (got:$PORT_NUMBER)"
	    exit 5;
	fi
	

	if [ $UID != 0 ]; then
	    echo "Error: needs to be executed with sudo (was executed by `id -u -n`)"
	    exit 6;
	fi;


#	echo "Will redirect $PLC_IP to port $PORT_NUMBER for project $PROJECT_NAME driver $DRIVER_NUMBER"

    } done

# Note that the above checks were done in a subshell; we need to re-throw the error to our shell
RETVAL=$?;
if [ $RETVAL -ne 0 ]; then exit $RETVAL;fi


    # then do it for real...

    # yet firstly clear all redirects
    if [ "a"${DO_CLEAR} != "a" ]; then 
#	echo "Clearing existing redirects"
	clearRedirects; 
    fi

cat $CFG_FILE|while read PLC_IP PORT_NUMBER PROJECT_NAME DRIVER_NUMBER;
    do {
	if [ -z $PLC_IP ]; then continue; fi; #skip empty lines
	if [ x${PLC_IP:0:1} == "x#" ]; then continue; fi; #skip comment lines
#	echo redirectPort $PLC_IP $PORT_NUMBER
	redirectPort $PLC_IP $PORT_NUMBER

    } done

RETVAL=$?;
if [ $RETVAL -ne 0 ]; then exit $RETVAL;fi

exit 0;

}

################# MAIN PROGRAM STARTS HERE #############
PAR1=
PAR2=
COMMAND=
PARAMS=

while [[ $# > 0 ]]; do
    case $1 in
    -configure|--configure)
        COMMAND=redirectFromCfgFile
        ;;
    -configureClear|--configureClear)
    echo this is configclear
        COMMAND=redirectFromCfgFile
        PARAMS=1
        ;;
    -clear|--clear)
	COMMAND=clearRedirects
        ;;
    -status|--status)
	COMMAND=showStatus;
        ;;
    -cfg|--cfg|-configFile|--configFile)
	shift
	CFG_FILE="$1"
	;;

    *) if [ x$PAR1 = "x" ]
       then
	PAR1=$1;
       else
        PAR2=$1;
       fi
       ;;
    esac
    shift
done


if [ x$COMMAND != "x" ]; then
    $COMMAND $PARAMS
    exit $?
fi

if [ x$PAR1 = "x" -o x$PAR2 = "x" ] ; then
    usage
    exit 1;
fi;

redirectPort $PAR1 $PAR2
