/**@file unCMWServer.ctl

This file contains commonly used functions for CMWServer component.
The main goal of functions in this file is providing common functionality
for CMWServer specific panels, plus also some very basic functions for
UNICOS system intergity support.<br>

This file also defines useful constants for interpreting the result of configuration
file parsing.
Parsing/validation of configuration file for server is implemented in CTRL DLL
CMWServerConfigValidator. This DLL contains two (main) functions:
<ul>
 <li> int CMWServerConfigParser_validateAll(string fileName, mapping &result, string &errMsg)</li>
 <li> int CMWServerConfigParser_validateErrorsOnly(string fileName, mapping &result, string &errMsg)</li>
</ul>
Both functions have the same set of arguments and returned value meaning, the difference
between two functions is: the first one returns in 'result' the whole content of configuration
file, while the second functions only return file entries which didn't pass validation.<br>
Arguments:
<ul>
 <li> fileName: the name of file to be parsed, the path is relative to project path,
                 for example: data/MyFile.config</li>
 <li> result: mapping where result of parsing will be written, see detailed description of
               mapping content below</li>
 <li> errMsg: variable where 'summary' error message will be written in case of errors</li>
</ul>
Both functions return 0 if parsing/validation was successful, or -1 if error was found.<br>
Probably, the most interesting/useful is mapping result. It contains a number of string keys
(see constants below), the value for each key is dyn_dyn_anytype. Length of listst is the same
for all keys in mapping. The number of elements in list is equal to the number of DPEs in
configuration file (for first function), or number of DPEs with errors (for second function).<br>
Each Nth element of list contains 3 values:
<ol>
 <li> string: the value of attribute as read from configuration file</li>
 <li> bool: true if this attribute passed validation, false otherwise</li>
 <li> string: detail error rescription if attribute didn't pass validation</li>
</ol>
*/

#uses "CMWServerConfigValidator"  // CTRL DLL: config file parsing, plus some auxiliary functions
#uses "unCMWServerDeprecated.ctl"  // Deprecatd functions are moved from here to that library

#uses "fwGeneral/fwException.ctl"
#uses "fwGeneral/fwExceptionHandling"
#uses "unMessageText"


/** @name The keys in mapping returned by config file parsing. */
//@{
const string CMWSERVER_CFG_KEY_CLASS = "class";          ///< Name of class for CMW device
const string CMWSERVER_CFG_KEY_DEVICE = "device";        ///< Name of CMW device
const string CMWSERVER_CFG_KEY_PROPERTY = "property";    ///< Name of property in CMW device
const string CMWSERVER_CFG_KEY_ALIAS = "alias";          ///< DPE/DP alias
const string CMWSERVER_CFG_KEY_DPE = "dpe";              ///< DPE name
const string CMWSERVER_CFG_KEY_DIRECTION = "direction";  ///< Data flow direction
const string CMWSERVER_CFG_KEY_VALUE_TYPE = "valueType"; ///< Data type of value tag
const string CMWSERVER_CFG_KEY_TIME_TYPE  = "timeType";  ///< Data type of time tag
const string CMWSERVER_CFG_KEY_LINE_NO = "lineNo";       ///< Line number(s) in configuration file
const string CMWSERVER_CFG_KEY_SYNTAX = "syntax";        ///< Syntax error(s) in block not related to particular item
const string CMWSERVER_CFG_KEY_COMMENT = "comment";      ///< Comment string in file
//@)

/** @name Indices of individual values in the list of CMWServer manager settings. */
//@{
const int CMWSERVER_MAN_CONFIGPATH = 1;    ///< Name of configuration file
const int CMWSERVER_MAN_NUMPERIOD = 2;     ///< Time interval for updating heartbeat DPE by CMWServer [msec]
const int CMWSERVER_CMW_SERVER = 3;        ///< Name of CMW server
const int CMWSERVER_CMW_DATATAGNAME = 4;   ///< Name of value tag
const int CMWSERVER_CMW_TIMETAGNAME = 5;   ///< Name of time tag
const int CMWSERVER_MAN_EV_TIMEOUT = 6;    ///< Time to wait EventManager response to dpSet() request [msec]
const int CMWSERVER_CMW_DISABLE_PARTIAL_SET = 7;  ///< If property contains both value and time tags, then only SET requests from client containing both value and time will be accepted
const int CMWSERVER_MAN_MIN_ALARM_TIME = 8;       ///< Minimum duration of alarm value before the value is allowed to change
//@)

/** @name Other constants */
//@{
/// Catalogue name
const string g_sCat_CMWServer="unCMWServer";

/// user allowed to modify the CMWServer if Unicos Package not installed
const string g_CMWServer_userAdmin="root"; 
//@)


/** @name Public functions */
//@(

/**
  Write default settings to configuration DP for CMWServer manager.
  Usually used when the DP for new manager is just added

  @param sDP The name of configuration DP for CMWServer manager
  @param exceptionInfo Errors description (if any) will be added th this variable

Usage: Public

PVSS manager usage: Ctrl, UI

Constraints:
	. PVSS version: 3.15
	. operating system: Linux, Windows.
	. distributed system: yes.
*/
void unCMWServer_DefaultSettingsManager(string sDP, dyn_string &exceptionInfo)
{
  int iTimeout;
  dyn_string dsSettings;
  string sFunction, sName;
  unCMWServer_Configuration(dsSettings, sFunction, sName, iTimeout, exceptionInfo);
  if (dynlen(exceptionInfo)==0)
  {
    if (dpExists(sDP) && (dynlen(dsSettings) >= CMWSERVER_MAN_MIN_ALARM_TIME))
    {
      int iRes = dpSet(
            sDP + ".Settings.man_ConfigPath", dsSettings[CMWSERVER_MAN_CONFIGPATH],
            sDP + ".Settings.man_NumPeriod",dsSettings[CMWSERVER_MAN_NUMPERIOD],
            sDP + ".Settings.cmw_Server", dsSettings[CMWSERVER_CMW_SERVER],
            sDP + ".Settings.cmw_DataTagName", dsSettings[CMWSERVER_CMW_DATATAGNAME],
            sDP + ".Settings.cmw_TimeTagName", dsSettings[CMWSERVER_CMW_TIMETAGNAME],
            sDP + ".Settings.EventManagerResponseTimeout", dsSettings[CMWSERVER_MAN_EV_TIMEOUT],
            sDP + ".Settings.DisablePartialSetting", dsSettings[CMWSERVER_CMW_DISABLE_PARTIAL_SET],
            sDP + ".Settings.MinAlarmDuration", dsSettings[CMWSERVER_MAN_MIN_ALARM_TIME],
            sDP + ".State.st_CommandIn", 100,				//Stop
            sDP + ".State.st_StateOut", 1000,			//Unknown state
            sDP + ".State.ev_MessageOut", c_Ok);		//OK
    }
    else
    {
      fwException_raise(exceptionInfo, "ERROR", __FUNCTION__ + "(): DP does not exist " + sDP, "");
    }
  }
}

/**
  Read the current settings for CMWServer manager from configuration DP

  @param sDP The name of configuration DP for CMWServer manager
  @param dsParam Settings from configuration DP will be written to this varible at
                 corresponding indices, see constants CMWSERVER_XXX
  @param exceptionInfo Errors description (if any) will be added th this variable
			
Usage: Public

PVSS manager usage: Ctrl, UI (WCCOANV) and UI extended (WCCOANG)

Constraints:
	. PVSS version: 3.15
	. operating system: Linux, Windows.
	. distributed system: yes.
*/
void unCMWServer_LoadSettingsManager(string sDP, dyn_string &dsParam, dyn_string &exceptionInfo)
{
  if (dpExists(sDP + ".Settings"))
  {
    int iRes = dpGet(
        sDP + ".Settings.man_ConfigPath", dsParam[CMWSERVER_MAN_CONFIGPATH],
        sDP + ".Settings.man_NumPeriod",dsParam[CMWSERVER_MAN_NUMPERIOD],
        sDP + ".Settings.cmw_Server", dsParam[CMWSERVER_CMW_SERVER],
        sDP + ".Settings.cmw_DataTagName", dsParam[CMWSERVER_CMW_DATATAGNAME],
        sDP + ".Settings.cmw_TimeTagName", dsParam[CMWSERVER_CMW_TIMETAGNAME],
        sDP + ".Settings.EventManagerResponseTimeout", dsParam[CMWSERVER_MAN_EV_TIMEOUT],
        sDP + ".Settings.DisablePartialSetting", dsParam[CMWSERVER_CMW_DISABLE_PARTIAL_SET],
        sDP + ".Settings.MinAlarmDuration", dsParam[CMWSERVER_MAN_MIN_ALARM_TIME]);
    if (iRes<0)
    {
      fwException_raise(exceptionInfo, "ERROR", __FUNCTION__ + "(): dpGet() failed for " + sDP + ".Settings.*", "");
    }
  }
  else
  {
    fwException_raise(exceptionInfo, "ERROR", __FUNCTION__ + "(): DP does not exists " + sDP + ".Settings", "");
  }
}

/**
  Get default settings for CMWServer manager. The default settings shall be stored
  in dedicated DPE when CMWServer component is installed. If by some reason (?) DP does
  not exist then this function creates such DP and writes there hardcoded defaults.

  @param dsSettings Variable where default settings for manager will be saved at appropriate
                    indices, see constants CMWSERVER_XXX
  @param sFunction_Name <i>Function Name to be used to update Table for configFile</i>. <b>Not used</b>,
                    remains in the list of argument in order to keep old signature
  @param sComponent_Name <i>Name of the Component used for this Manager</i>. <b>Not used</b>,
                    remains in the list of argument in order to keep old signature
  @param iIntegrity_Timeout Timeout to check if manager alive [sec]. Not used in practice,
                    remains in the list of argument in order to keep old signature
  @param exceptionInfo Errors description (if any) will be added th this variable
						
Usage: Public

PVSS manager usage: UI (WCCOANV) and UI extended (WCCOANG)

Constraints:
	. PVSS version: 3.0 
	. operating system: W2000, NT, XP and Linux, but tested only under XP and Linux.
	. distributed system: yes.
*/
void unCMWServer_Configuration(dyn_string &dsSettings, string &sFunction_Name, string &sComponent_Name, int &iIntegrity_Timeout, dyn_string &exceptionInfo)
{
  string sSystemName = getSystemName();
  if (dpExists(sSystemName + "CMWServer_Info.defaultSettings"))
  {
    int iRes = dpGet(sSystemName + "CMWServer_Info.defaultSettings.Settings", dsSettings);
    if(iRes < 0)
    {
      fwException_raise(exceptionInfo, "ERROR", __FUNCTION__ + "(): dpGet() failed for " + sSystemName + "CMWServer_Info.defaultSettings", "");
    }
    while(dynlen(dsSettings) < CMWSERVER_MAN_MIN_ALARM_TIME)
    {
      dynAppend(dsSettings, "");
    }
  }
  else
  {
    // Hardcoded default settings
    iIntegrity_Timeout = 300;
    dsSettings[CMWSERVER_MAN_CONFIGPATH] = "data/CMWServerData";
    dsSettings[CMWSERVER_MAN_NUMPERIOD] = "30000";
    dsSettings[CMWSERVER_CMW_SERVER] = "CMWServerName";
    dsSettings[CMWSERVER_CMW_DATATAGNAME] = "value";
    dsSettings[CMWSERVER_CMW_TIMETAGNAME] = "time";
    dsSettings[CMWSERVER_MAN_EV_TIMEOUT] = "30000";
    dsSettings[CMWSERVER_CMW_DISABLE_PARTIAL_SET] = "false";
    dsSettings[CMWSERVER_MAN_MIN_ALARM_TIME] = "1000"; 

    // Create DP for default settings, write
    int iRes = dpCreate("CMWServer_Info", "_Un_CMWServer_Config");
    if(iRes < 0)
    {
      fwException_raise(exceptionInfo, "ERROR", __FUNCTION__ + "(): dpCreate(CMWServer_Info) failed", "");
    }
    else
    {
      iRes = dpSetWait(sSystemName + "CMWServer_Info.defaultSettings.Settings", dsSettings);
      if(iRes < 0)
      {
        fwException_raise(exceptionInfo, "ERROR", __FUNCTION__ + "(): dpSet(CMWServer_Info.defaultSettings.Settings) failed", "");
      }
    }
  }
}

/**
  Display exception, the same piece of code is used in many CMWServer-related panels

  @param exceptionInfo Exception information to be displayed

Usage: Internal

PVSS manager usage: UI, CTRL

Constraints:
	. PVSS version: 3.15
	. operating system: Windows, Linux
	. distributed system: yes.
*/
void unCMWServer_displayException(dyn_string exceptionInfo)
{
  if(dynlen(exceptionInfo) == 0)
  {
    return;
  }
  if(isFunctionDefined("unMessageText_sendException"))
  {
    unMessageText_sendException("*", "*", "CMWServer", "user", "*", exceptionInfo);
  }
  else if(myManType() == UI_MAN)
  {
    fwExceptionHandling_display(exceptionInfo);
  }
  else
  {
    DebugTN(__FUNCTION__, exceptionInfo);
  }
}

/**
Decode manager number from config DP name: last part of DP name after '_' is converted
to integer; it is expected that last part must be "01"..."99"

  @param sManagerConfigDp CMWServer configuration DP name
  @param exceptionInfo Errors description (if any) will be added th this variable

  @return Last part of DP name converted to integer

Usage: Internal

PVSS manager usage: UI, CTRL

Constraints:
  . PVSS version: 3.15
  . operating system: Windows, Linux
  . distributed system: yes.
*/
int unCMWServer_decodeManNum(const string &sManagerConfigDp, dyn_string &exceptionInfo)
{
  int iResult = 0;
  dyn_string dsStrSplit = strsplit(dpSubStr(sManagerConfigDp, DPSUB_DP), "_");
  if(dynlen(dsStrSplit) > 1)  // Must be at least two parts
  {
    iResult = dsStrSplit[dynlen(dsStrSplit)];
  }
  if(iResult <= 0)
  {
    fwException_raise(exceptionInfo, "ERROR", __FUNCTION__ + ": wrong manager number " + iResult +
                     " is found from DP name " + sManagerConfigDp, "");
  }
  return iResult;
}

//@}
