/** @file unSystemIntegrity_CMWServer.ctl

  The functions for ineractions between CMWServer and UNICOS system integrity components.
*/

#uses "unicos_declarations.ctl"
#uses "unMessageText.ctl"
#uses "unGenericFunctions_core.ctl"
#uses "unCoreLHCServices.ctl"
#uses "unSystemIntegrity.ctl"
#uses "unAlarmConfig.ctl"
#uses "unCMWServerStatus.ctl"  // Calculation of alarm status

/** @name Constants */
//@(

/// The base part of system integrity DP for CMWServer
private const string UN_SYSTEM_INTEGRITY_CMWServer = "CMWServer";
//@}

//@{

//------------------------------------------------------------------------------------------------------------------------
// CMWServer_systemIntegrityInfo
/** Return the list of systemAlarm pattern, in this case - one pattern only.
  
@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  CTRL

@return list of systemAlarm pattern
*/
dyn_string CMWServer_systemIntegrityInfo()
{
  return makeDynString(CMWServer_pattern);
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_CMWServer_Initialize
/**
Get the list of CMWServer check by the systemIntegrity. The ones that are enabled.

@param dsResult the list of enabled CMWServer which are checked is returned here

Usage: Public

PVSS manager usage: CTRL (WCCOACTRL)

Constraints:
	. PVSS version: 2.12.1 
	. operating system: W2000, NT and Linux, but tested only under W2000 and Linux.
	. distributed system: yes.
*/
void unSystemIntegrity_CMWServer_Initialize(dyn_string &dsResult)
{
  unCoreLHCServices_initialize(c_CMWServer_Pattern);  // const in unCoreLHCServices.ctl
	
  // get the list of CMWServer manager dp to checked
  dyn_string dsList = dpNames(c_unSystemAlarm_dpPattern + CMWServer_pattern + "*", c_unSystemAlarm_dpType);
  int len = dynlen(dsList);
  for(int i = 1; i <= len;  i++)
  {
    string dpToCheck = dpSubStr(dsList[i], DPSUB_DP);
    dpToCheck = substr(dpToCheck, strpos(dpToCheck, CMWServer_pattern));  // const in unicos_declarations.ctl
    bool bEnabled;  // If system integrity check is enabled for this manager 
    dpGet(dsList[i] + ".enabled", bEnabled);
    if(bEnabled)
    {
      dynAppend(dsResult, dpToCheck);
    }
  }
  string sMessage = unGenericDpFunctions_getComponentsVersion(makeDynString("unCMWServer"));
  dyn_string exceptionInfo;
  unMessageText_send("*", "*", g_SystemIntegrity_sSystemName + "systemIntegrity", "user", "*", "INFO",
                     "systemIntegrity " + sMessage + " loaded", exceptionInfo);
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_CMWServer_HandleCommand
/**
Purpose:
Handle any systemIntegrity command. The commands are usually issued by user in some
panel of UI, but they are processed by CTL manager. The CTRL manager receives commands
in DPEs (CTRL receives callbacks), this function processes commands for system integhrity
of CMWServer manager.

@param sDpe1 The name of DPE where command came from (?), not used here
@param command The systemIntegrity command
@param sDpe2 The name of DPE where command parameters came from (?), not usd here
@param parameters The list of parameters of the systemIntegrity command

Usage: Public

PVSS manager usage: CTRL (WCCOACTRL)

Constraints:
	. PVSS version: 2.12.1 
	. operating system: W2000, NT and Linux, but tested only under W2000 and Linux.
	. distributed system: yes.
*/
void unSystemIntegrity_CMWServer_HandleCommand(string sDpe1, int command, string sDpe2, dyn_string parameters)
{
  dyn_string exceptionInfo;
  int i, len = dynlen(parameters);
  string sMessage;

  switch(command)
  {
    case UN_SYSTEMINTEGRITY_ADD: // add the _UnSystemAlarm DP
      for(i = 1 ; i <= len ; i++)
      {
        unSystemIntegrity_CMWServer_checking(parameters[i], true, true, exceptionInfo);
      }
      break;
    case UN_SYSTEMINTEGRITY_DELETE: // delete the _UnSystemAlarm DP
      for(i = 1 ; i <= len ; i++)
      {
        unSystemIntegrity_CMWServer_checking(parameters[i], false, false, exceptionInfo);
        // remove dp from the alert list of applicationDP if it is in
        _unSystemIntegrity_modifyApplicationAlertList(c_unSystemAlarm_dpPattern + parameters[i], false, exceptionInfo);
        // delete the _UnSystemAlarm dps
        if(dpExists(c_unSystemAlarm_dpPattern + parameters[i]))
        {
          dpDelete(c_unSystemAlarm_dpPattern + parameters[i]);
        }
      }
      break;
    case UN_SYSTEMINTEGRITY_ENABLE: // enable the _UnSystemAlarm DP
      for(i = 1 ; i <= len ; i++)
      {
        if(dpExists(c_unSystemAlarm_dpPattern + parameters[i]))
        {
          unSystemIntegrity_CMWServer_checking(parameters[i], false, true, exceptionInfo);
        }
      }
      break;
    case UN_SYSTEMINTEGRITY_DISABLE: // disable the _UnSystemAlarm DP
      for(i = 1 ; i <= len ; i++)
      {
        if(dpExists(c_unSystemAlarm_dpPattern + parameters[i]))
        {
          unSystemIntegrity_CMWServer_checking(parameters[i], false, false, exceptionInfo);
        }
      }
      break;
    case UN_SYSTEMINTEGRITY_DIAGNOSTIC: // give the list of _UnSystemAlarm DP and the state
      dpSet(UN_SYSTEM_INTEGRITY_CMWServer + UN_SYSTEMINTEGRITY_EXTENSION + ".diagnostic.commandResult", command,
            UN_SYSTEM_INTEGRITY_CMWServer + UN_SYSTEMINTEGRITY_EXTENSION + ".diagnostic.result", CMWServerStatus::getAllAlarms());
      break;
    case UN_SYSTEMINTEGRITY_VERSION:
      sMessage = unGenericDpFunctions_getComponentsVersion(makeDynString("unCMWServer"));
      dpSet(UN_SYSTEM_INTEGRITY_CMWServer + UN_SYSTEMINTEGRITY_EXTENSION + ".diagnostic.commandResult", command, 
            UN_SYSTEM_INTEGRITY_CMWServer + UN_SYSTEMINTEGRITY_EXTENSION + ".diagnostic.result", makeDynString(getCurrentTime(),
                                                                                                               sMessage));
      break;
    default:
      break;
  }

  if(dynlen(exceptionInfo) > 0)
  {
    if(isFunctionDefined("unMessageText_sendException"))
    {
      unMessageText_sendException("*", "*", "unSystemIntegrity_CMWServer_HandleCommand", "user", "*", exceptionInfo);
    }
    // handle any error uin case the send message failed
    if(dynlen(exceptionInfo) > 0)
    {
      DebugTN(__FUNCTION__, exceptionInfo);
    }
  }
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_CMWServer_checking
/**
Purpose:
This function register/de-register the callback function of CMWServer manager. This function can also create the _unSystemAlarm_CMWServer dp 
with the alarm config but it cannot delete it.

  @param sTypeAndMum manager type and name, for exmple, "CMWServer_5" for CMWServer manager #5
  @param bCreate true create the dp and the alarm config if it does not exist, enable it
  @param bRegister true do a dpConnect, false do a dpDisconnect
  @param exceptionInfo exception are returned here

Usage: Internal

PVSS manager usage: CTRL (WCCOACTRL)

Constraints:
	. PVSS version: 2.12.1 
	. operating system: W2000, NT and Linux, but tested only under W2000 and Linux.
	. distributed system: yes.
*/

void unSystemIntegrity_CMWServer_checking(string sTypeAndMum, bool bCreate, bool bRegister, dyn_string &exceptionInfo)
{
  string sSystem = getSystemName();
  
  //1. Get Alarm Name
  string sAlarmDp = c_unSystemAlarm_dpPattern + sTypeAndMum;
	
  //2. Get Manager config DP name
  string sManNum = substr(sTypeAndMum, strpos(sTypeAndMum, "_") + 1);
  string sManagerDp = CMWServer_pattern + "Manager_" + (strlen(sManNum) < 2 ? "0" : "") + sManNum;

  //3. Connect/disconnect
  dyn_errClass err;
  if(bRegister)
  {
    //Create the sAlarmDp if it is not existing
    if(!dpExists(sAlarmDp))
    {
      string description = getCatStr("unSystemIntegrity", "CMWServer_DESCRIPTION") + sManNum;
      unSystemIntegrity_createSystemAlarm(sManNum, CMWServer_pattern, description, exceptionInfo);
    }
    if(CMWServerStatus::startManagerMonitoring(sSystem, sManagerDp, sAlarmDp, exceptionInfo))
    {
      // enable alarm if monitorin started successfuly, the same functionality as in _unSystemIntegrity_registerFunction()
      dpSet(sAlarmDp + ".enabled", true);
      unAlarmConfig_mask(sAlarmDp + ".alarm", false, exceptionInfo);
    }
  }
  else
  {
    CMWServerStatus::stopManagerMonitoring(sManagerDp, exceptionInfo);
    dpSet(sAlarmDp + ".enabled", false,
          sAlarmDp + ".alarm", c_unSystemIntegrity_no_alarm_value);
    unAlarmConfig_mask(sAlarmDp + ".alarm", true, exceptionInfo);
  }
}

//@}
