#uses "fwGeneral/fwDPELock.ctl"
#uses "classes/unModbusRedirect/UnModbusRedirectConfiguration.ctl"

class UnModbusRedirectServer {

    protected static const string defaultDPName = "unModbusRedirect";

	// these are initialized from DPs (or defaults are set if empty)
	// by a dpConnect initial trigger
	protected string cfgFile;
	protected string redirectModbusCmd; 
	
	protected string cfgFileError;
	protected string cmdError;
    protected string dp;
	protected bool initialized;

    private UnModbusRedirectServer(string mDpName=defaultDPName) 
    {
		dp = mDpName;
		initialized = false;
    }

    public static shared_ptr<UnModbusRedirectServer> getInstance() 
    {
		if (equalPtr(m_instance,nullptr)) {
	    	m_instance = new UnModbusRedirectServer();
	    	m_instance.initialize();
		}
		return m_instance;
    }

    private static shared_ptr<UnModbusRedirectServer> m_instance=nullptr;

    protected void commandCB(string dpe, string value)
    {

		if (!initialized) {
			throwError(makeError("unModbusRedirect", PRIO_WARNING, ERR_CONTROL, 2000,"Server received command but not yet initialized - ignoring: ",value));
			return;
		}

		// value is a string; if it needs to have parameters,
		// then there is a "#" separator at the end of command,
		// and after that the parameters are separated by a "|"
	
		string cmd=value;
		dyn_string params;  // either strsplit'ed
		string paramstring; // or as one string, e.g. serialized

		int idx=uniStrPos(value,"#");
		if (idx>=0) {
	    	cmd=uniSubStr(value,0,idx);
	    	if (uniStrLen(value)>idx) {
				paramstring=uniSubStr(value,idx+1);
				params=strsplit(paramstring,"|");
	    	}
		}
		switch (cmd) {
	    	case "PING":
				SendResponse("OK");
				break;
	    	case "GETHOSTINFO":
				SendResponse(GetInfo());
				break;
	    	case "CHECKCMD":
				SendResponse(CheckCommandErrors());
				break;
	    	case "GETCONFIGURATION":
				SendResponse(GetConfiguration());
				break;
	    	case "SETCONFIGURATION":
				SendResponse(SetConfiguration(paramstring));
				break;
	    	case "CONFIGURE":
				SendResponse(Configure(false));
				break;
	    	case "CONFIGURECLEAR":
				SendResponse(Configure(true));
				break;
	    	case "CLEAR":
				SendResponse(Clear());
				break;
	    	default:
	    		throwError(makeError("unModbusRedirect", PRIO_WARNING, ERR_CONTROL, 2000,"Server received unrecognized command: ",value));
				SendResponse("ERROR#WRONG_COMMAND "+cmd);
				break;
		}
    }

	protected string CheckCommandErrors()
	{
		if (cmdError!="") return "ERROR#"+cmdError;
	}

	protected UnModbusRedirectConfiguration getCurrentConfiguration()
	{
		UnModbusRedirectConfiguration retcfg;
		// assumes that the checking for the command, etc was already done...
		string cmd="sudo "+redirectModbusCmd+" -status";
		string out,err;
		int rc=system(cmd,out,err);
		if (rc!=0) return nullptr;
		dyn_string ds=strsplit(out,"\n");
		for (int i=1;i<=dynlen(ds);i++) {
			string s=strltrim(strrtrim(ds[i]));
			if (s=="") continue;
			string plcip;
			int port;
			int numTokens=sscanf(s,"%s => %d\n",plcip, port);
			if (numTokens!=2) {
				throwError(makeError("unModbusRedirect", PRIO_WARNING, ERR_CONTROL, 2000,"Error parsing current configuration ( "+numTokens+") in line "+i+"(skipping): "+s));
				continue;
			}
			UnModbusRedirectEntry e=UnModbusRedirectEntry(plcip, port);
			e.conf_live=true;
			retcfg.addEntry(e);
		}
		return retcfg;
	}

	protected string GetInfo()
	{
		return PROJ+"|"+getHostname();
	}

	protected string Configure(bool doClear)
	{
		if (redirectModbusCmd=="") return "ERROR#Configure: RedirectModbus command invalid";
		if (cfgFileError!="") return "ERROR#Configure: "+cfgFileError;
		if (cfgFile=="") return "ERROR#Configure:NoConfig file";

		string cmd="sudo "+redirectModbusCmd;
		if (doClear) cmd+=" -configureClear"; else cmd+=" -configure";
		cmd+=" -cfg "+cfgFile;
		
		string out,err;
		int rc=system(cmd,out,err);
		if (rc) return "ERROR#Configure: "+out+";"+err;
	
		return "OK";
	}

	protected string Clear()
	{
		if (redirectModbusCmd=="") return "ERROR#Clear: RedirectModbus command invalid";

		string cmd="sudo "+redirectModbusCmd+" -clear";
		
		string out,err;
		int rc=system(cmd,out,err);
		if (rc) return "ERROR#Clear: "+out+";"+err;
	
		return "OK";
	}

    protected string GetConfiguration()
    {
		if (cfgFileError!="") return "ERROR#GetConfiguration: "+cfgFileError;

		// get one from file
		UnModbusRedirectConfiguration confFile=UnModbusRedirectConfiguration::fromFile(cfgFile);
		if (confFile.error!="") return "ERROR#GetConfiguration: "+confFile.error;
		
		if (redirectModbusCmd=="") return "ERROR#GetConfiguration: RedirectModbus command invalid";
		UnModbusRedirectConfiguration confLive = getCurrentConfiguration();
		if (confLive.error!="") return "ERROR#GetConfiguration: "+confLive.error;

		// this will be the effective one
		UnModbusRedirectConfiguration conf;
	
		// combine the two; untill we have good collections, we need to do it manually...
		for (int i=0;i<confFile.numEntries();i++) { //zero-based iteration with at()!
			UnModbusRedirectEntry e=confFile.at(i);
			int idx=confLive.find(e.plcIP,e.port);
			if (idx>=0) {
				// found:
				e.conf_live=true;
			} else {
				e.conf_live=false;
			}
			conf.addEntry(e);
//DebugTN("WE'D NEED SHARED PTRS HERE");
		}
		// now look for those that are live but not in the file...
		for (int i=0;i<confLive.numEntries();i++) { //zero-based iteration with at()!
			UnModbusRedirectEntry e=confLive.at(i);
			int idx=confFile.find(e.plcIP,e.port);
			if (idx<0) {
				e.conf_inFile=false;
				conf.addEntry(e);
			}
		}
		return conf.serialize();
    }

	protected string SetConfiguration(string data)
	{
		
		UnModbusRedirectConfiguration cfg=UnModbusRedirectConfiguration::deserialize(data);
		cfg.saveToFile(cfgFile);
		if (cfg.error!="") return "ERROR#SetConfiguration: "+cfg.error;

		return "#OK";
	}

    protected void clientLockCB(string dpe, bool locked)
    {
		dyn_string lockDetails, exceptionInfo;
	
		if (!locked) return;

		string whoLocks="";
		fwDPELock_getLocked(dpSubStr(dpe,DPSUB_SYS_DP_EL), lockDetails, exceptionInfo);
		if (!dynlen(exceptionInfo)) {
	    	whoLocks = lockDetails[fwDPELock_LOCK_USER_NAME]       + "@" +
					   lockDetails[fwDPELock_LOCK_MANAGER_DETAIL]  + " #"+
		       		   lockDetails[fwDPELock_LOCK_MANAGER_REPLICA] + " ("+
		       		   lockDetails[fwDPELock_LOCK_MANAGER_HOST]    + ")";
		}
    }


    protected synchronized void configFileChangedCB(string dpe, string value)
    {
    
    	string origCfgFile=cfgFile;
		cfgFileError=""; // reset the previous error
		bool createFile=false;
		
		if (value=="*DEFAULT*") {
			value=dirName(dirName(PROJ_PATH))+"redirectModbus.cfg";
			dpSetWait(dp+".configFile",value); // reset in the DP; for this reason we need synchronized callback
		}

		if (value=="*CREATEDEFAULT*") {
			value=dirName(dirName(PROJ_PATH))+"redirectModbus.cfg";
			createFile=true;
			dpSetWait(dp+".configFile",value); // reset in the DP; for this reason we need synchronized callback
		}

		if (cfgFile==value && !createFile) return;


    	cfgFile=value;

		if (cfgFile=="" ) {
			// only retrigger callbacks and do not proceed...
			dpSetWait(dp+".configFile",cfgFile);
			return;
		}

    	if (initialized) {
    		// this is a change rather than initialization: report it
			throwError(makeError("unModbusRedirect", PRIO_INFO, ERR_CONTROL, 2000,"Configuration file change request to",value));
		}

		// check if file exists
		int rc = access(cfgFile, R_OK);
		
		if (rc!=0 && !createFile) {
			cfgFileError="Configuration file does not exist.";
			throwError(makeError("unModbusRedirect", PRIO_WARNING, ERR_CONTROL, 2000,"Config file does not exist: "+cfgFile));
			cfgFile="";
			dpSetWait(dp+".configFile:_original.._exp_inv",true);
			return;
		}
		
		// note: if it exists and it has zero size we also need to recreate it
		if ( (rc==0) && (getFileSize(cfgFile)>0)) createFile=false;
		
		if (createFile) {
			throwError(makeError("unModbusRedirect", PRIO_WARNING, ERR_CONTROL, 2000,"Creating new config file "+cfgFile));
			string templFile=getPath(DATA_REL_PATH,"unModbusRedirect/redirectModbus.cfg.template");
			if (templFile=="") {
				cfgFileError="Template for configuration file not found.";
				throwError(makeError("unModbusRedirect", PRIO_WARNING, ERR_CONTROL, 2000,cfgFileError));
				cfgFile="";
				dpSetWait(dp+".configFile",templFile,
					      dp+".configFile:_original.._exp_inv",true);
				return;
			}
			bool ok=copyFile (templFile,cfgFile);
			if (!ok) {
				cfgFileError="Could not create a default config file.";
				throwError(makeError("unModbusRedirect", PRIO_WARNING, ERR_CONTROL, 2000,cfgFileError));
				cfgFile="";
				dpSetWait(dp+".configFile",templFile,
					      dp+".configFile:_original.._exp_inv",true);
				return;
			}
		}

		string s;
		bool ok=fileToString(cfgFile,s);
		if (!ok) {
			cfgFileError="Configuration file not readable.";
			throwError(makeError("unModbusRedirect", PRIO_SEVERE, ERR_CONTROL, 2000,"Config file not readable: "+cfgFile));
			cfgFile="";
			dpSetWait(dp+".configFile:_original.._exp_inv",true);
			return;
		}
		
		if (cfgFile!=value) {
			// means that this was modified by us somehow,
			// eg. set to default, and in addition it was succesful;
			// set to the datapoint
			dpSetWait(dp+".configFile",cfgFile, dp+".configFile:_original.._exp_inv",false);
		}
    }

	protected void redirectModbusCommandChangedCB(string dpe, string cmd)
	{
		string origCmd=cmd; // preserve
		string out, err;
		int rc;
		cmdError=""; // reset the error
		string prevCmd=redirectModbusCmd; // preserve for some comparisons
		redirectModbusCmd=""; // and reset the new one

    	if (cmd!="") {
    		if (cmd==redirectModbusCmd) return;
    		
    		if (initialized) {
    			// report the change
    			throwError(makeError("unModbusRedirect", PRIO_INFO, ERR_CONTROL, 2000,"redirectModbus command changed to "+cmd));
    		}
			rc=access(cmd,X_OK);
			if (rc) {
				cmdError="redirectModbus command not found or not executable";
    			throwError(makeError("unModbusRedirect", PRIO_WARNING, ERR_CONTROL, 2000,cmdError,cmd));
    			dpSetWait(dp+".redirectModbusCmd:_original.._exp_inv",true);
    			return;
			}

    		rc=system("sudo -l "+cmd,out,err);
    		if (rc) {
    			cmdError="redirectModbus command not available for execution with sudo";
    			throwError(makeError("unModbusRedirect", PRIO_WARNING, ERR_CONTROL, 2000,cmdError));
    			dpSetWait(dp+".redirectModbusCmd:_original.._exp_inv",true);
    			return;
    		}
			
    		redirectModbusCmd=cmd;
    		if (prevCmd=="") {
    			// means it was empty or invalid => reset the invalid bit
    			dpSetWait(dp+".redirectModbusCmd:_original.._exp_inv",false);
    		}
    		return;
    	}

		throwError(makeError("unModbusRedirect", PRIO_INFO, ERR_CONTROL, 2000,"command not specified: will try to guess one"));
		    	
    	// otherwise, if not set then try to guess one
    	// get the one that is declared explicily in the sudoers file	
		rc=system("sudo -ln |awk '/redirectModbus.sh/{print $3;exit 1}'",out,err);
		if (rc==1) {
			cmd=strrtrim(out);
			if (access(cmd,X_OK)==0) {
    			throwError(makeError("unModbusRedirect", PRIO_INFO, ERR_CONTROL, 2000,"redirectModbus will use command configured explicitely in sudo:",cmd));
    			dpSetWait(dp+".redirectModbusCmd",cmd,
    			          dp+".redirectModbusCmd:_original.._exp_inv",false); // this will retrigger us (as well as the client), hence we may simply return
				return;
			} else {
    			throwError(makeError("unModbusRedirect", PRIO_WARNING, ERR_CONTROL, 2000,"redirectModbus cannot use command configured explicitely in sudo (not executable):",cmd));
			}
		} 
		
		// otherwise, maybe there is a global sudo rule permitting to use the one
		// installed in the project
				
		string localModbusRedirect=getPath(DATA_REL_PATH,"unModbusRedirect/redirectModbus.sh");
		rc=system("sudo -l "+localModbusRedirect,out, err);
		if (rc==0) {
			cmd=strrtrim(out);
			if (access(cmd,X_OK)==0) {
    			throwError(makeError("unModbusRedirect", PRIO_INFO, ERR_CONTROL, 2000,"redirectModbus will use the script from the project:",cmd));
    			dpSetWait(dp+".redirectModbusCmd",cmd,
    			          dp+".redirectModbusCmd:_original.._exp_inv",false); // this will retrigger us (as well as the client), hence we may simply return
				return;
			} else {
    			throwError(makeError("unModbusRedirect", PRIO_WARNING, ERR_CONTROL, 2000,"redirectModbus cannot use the script from the project (not executable):",cmd));
			}
		} 

		if (cmd!=origCmd) {
			// means that this was modified by us somehow,
			// eg. set to default, and in addition it was succesful;
			// set to the datapoint to let the client know...
			dpSetWait(dp+".redirectModbusCmd",cmd,
			          dp+".redirectModbusCmd:_original.._exp_inv",false);
		}


	}

	/*
		Special sequence for initialization signaling, so that we could wait for all our own
		callbacks to settle down.
		
		1) We lock the ".response"
		1.5) Note that we have a transient state here where we re-locked and may still have the old "STARTED" value 
		     from the previous execution! Hence:
		2) We "pulse" the response with the "STARTING" value (clients should see it)
		3) We wait for all callbacks, etc
		4) We send the "STARTED" as response
	**/
    private void initialize() 
    {
		dyn_string exceptionInfo;
	
		// Wait for a lock, reporting every 5 minutes (300 seconds)
		while (true) {
	    	bool gotTheLock=fwDPELock_tryLock(dp+".response", exceptionInfo, 300, 0);
	    	if (gotTheLock) break; 
	    	if (dynlen(exceptionInfo)) {
	    		throwError(makeError("unModbusRedirect", PRIO_WARNING, ERR_CONTROL, 2000,"Server initialize error: cannot get the lock",exceptionInfo[2]));
	    	}
		}
		// 1.5): transient state; to be able to resolve it, we wait a bit (to get callbacks on the client side ordered correctly, 
		//       as there is no atomicity that we could assume)
		// and then 2): send the STARTING pulse
		delay(0,100);
		SendResponse("STARTING");
		// do all the initialization
		dpConnect(m_instance, m_instance.commandCB,FALSE,dp+".command");
		dpConnect(m_instance, m_instance.clientLockCB,FALSE,_fwDPELock_getLockConfig(dp+".command")+"._locked");
		dpConnect(m_instance, m_instance.configFileChangedCB,TRUE, dp+".configFile");
		dpConnect(m_instance, m_instance.redirectModbusCommandChangedCB,TRUE, dp+".redirectModbusCmd");
		
		// we should wait for initial callbacks to trigger here before sending the confirmation that we started
		// we do it in a trivial way by a simple delay - we could check individual initializations instead...
		delay(0,500);
		throwError(makeError("unModbusRedirect", PRIO_INFO, ERR_CONTROL, 2000, "Server initialized OK."));
		initialized = true;
		SendResponse("STARTED");
    }

    protected void SendResponse(string response)
    {
		dpSetWait(dp+".response",response);
    }
};
